@@ -39,6 +39,7 @@
 #include <libxml/xpath.h>
 #include <libxml/xpathInternals.h>
 #include <libxml/xmlwriter.h>
+#include <libgnome/gnome-config.h>
 
 #ifdef YELP_DEBUG
 #define d(x) x
@@ -52,12 +53,20 @@
 #define COL_HEADER 3
 #define TOC_PATH "ui/menubar/BookmarksMenu/BookmarksTOC"
 #define DOC_PATH "ui/menubar/BookmarksMenu/BookmarksDOC"
+#define BK_CONFIG_WIDTH  "/yelp/Bookmarks/width"
+#define BK_CONFIG_HEIGHT "/yelp/Bookmarks/height"
+#define BK_CONFIG_WIDTH_DEFAULT  "360"
+#define BK_CONFIG_HEIGHT_DEFAULT "360"
 
 static GSList *windows;
 static GtkTreeStore *actions_store;
 static gchar *dup_title;
 
 static GtkWidget *bookmarks_dialog = NULL;
+static GtkWidget *edit_open_button = NULL;
+static GtkWidget *edit_rename_button = NULL;
+static GtkWidget *edit_remove_button = NULL;
+
 
 static gboolean have_toc = FALSE;
 static gboolean have_doc = FALSE;
@@ -72,6 +81,16 @@ struct _YelpWindowData {
     GtkActionGroup *doc_actions;
 };
 
+static const char ui_description[] =
+    "<ui>"
+    "  <popup>"
+    "    <menuitem action='OpenName'/>"
+    "    <separator/>"
+    "    <menuitem action='EditName'/>"
+    "    <menuitem action='DeleteName'/>"
+    "  </popup>"
+    "</ui>";
+
 static void      window_add_bookmark        (YelpWindowData *window,
 					     gchar          *name,
 					     const gchar    *label);
@@ -100,7 +119,7 @@ static void      bookmarks_rename_button_cb (GtkWidget         *widget,
 					     GtkTreeView       *view);
 static void      bookmarks_remove_button_cb (GtkWidget         *widget,
 					     GtkTreeView       *view);
-
+static void      bookmarks_ensure_valid     (void);
 static gboolean  bookmarks_read             (void);
 
 static gboolean  bookmarks_dup_finder       (GtkTreeModel   *model,
@@ -115,6 +134,37 @@ static void      bookmarks_cell_edited_cb   (GtkCellRendererText *cell,
 					     const gchar         *path_string,
 					     const gchar         *new_title,
 					     gpointer             user_data);
+static void      bookmarks_popup_menu       (GtkWidget           *widget,
+					     GdkEventButton      *event);
+static void      bookmarks_menu_edit_cb     (GtkAction           *action,
+					     GtkWidget           *widget);
+static void      bookmarks_menu_remove_cb   (GtkAction           *action,
+					     GtkWidget           *widget);
+static void      bookmarks_menu_open_cb     (GtkAction           *action,
+					     GtkWidget           *widget);
+static gboolean  bookmarks_button_press_cb  (GtkWidget           *widget,
+					     GdkEventButton      *event);
+static gboolean  bookmarks_configure_cb     (GtkWidget           *widget, 
+					     GdkEventConfigure   *event,
+					     gpointer             data);
+static void      selection_changed_cb       (GtkTreeSelection    *selection,
+					     gpointer             data);
+
+static const GtkActionEntry popup_entries[] = {
+    { "OpenName", GTK_STOCK_OPEN,
+      N_("Open Bookmark in New Window"),
+      NULL, NULL,
+      G_CALLBACK (bookmarks_menu_open_cb)},
+    { "EditName", NULL,
+      N_("Rename Bookmark"),
+      NULL, NULL,
+      G_CALLBACK (bookmarks_menu_edit_cb)},
+    { "DeleteName", GTK_STOCK_REMOVE,
+      N_("Remove Bookmark"),
+      NULL, NULL,
+      G_CALLBACK (bookmarks_menu_remove_cb)}
+};
+
 
 void
 yelp_bookmarks_init (void)
@@ -190,6 +240,21 @@ yelp_bookmarks_register (YelpWindow *window)
     windows = g_slist_append (windows, data);
 }
 
+static void
+bookmarks_ensure_valid (void)
+{
+	if (!gtk_tree_model_iter_has_child (GTK_TREE_MODEL (actions_store),
+					    &toc_iter)) {
+	    gtk_tree_store_remove (actions_store, &toc_iter);
+	    have_toc = FALSE;
+	}
+	if (!gtk_tree_model_iter_has_child (GTK_TREE_MODEL (actions_store),
+					    &doc_iter)) {
+	    gtk_tree_store_remove (actions_store, &doc_iter);
+	    have_doc = FALSE;
+	}
+}
+
 static gboolean 
 bookmarks_dup_finder (GtkTreeModel *model, GtkTreePath *path,
 		      GtkTreeIter *iter, gpointer data)
@@ -223,7 +288,7 @@ yelp_bookmarks_add (const gchar *uri, YelpWindow *window)
     gchar       *title;
     gchar       *dup_uri;
 
-    title = gtk_window_get_title (GTK_WINDOW (window));
+    title = (gchar *) gtk_window_get_title (GTK_WINDOW (window));
     dup_uri = g_strdup (uri);
     
     if (dup_title)
@@ -360,7 +425,8 @@ bookmarks_add_bookmark (const gchar  *uri,
 			-1);
     if (save) {
 	for (cur = windows; cur != NULL; cur = cur->next) {
-	    window_add_bookmark ((YelpWindowData *) cur->data, uri, title);
+	    window_add_bookmark ((YelpWindowData *) cur->data, (gchar *) uri, 
+				 title);
 	}
 	yelp_bookmarks_write ();
     }
@@ -451,31 +517,32 @@ bookmarks_key_event_cb (GtkWidget   *widget,
     model = gtk_tree_view_get_model (view);
     sel = gtk_tree_view_get_selection (view);
     
-    gtk_tree_selection_get_selected (sel, &model, &iter);
-    path = gtk_tree_model_get_path (model, &iter);
-
-    switch (event->keyval) {
-    case GDK_BackSpace:
-    case GDK_Delete:
-	gtk_tree_store_remove (actions_store, &iter);
-	bookmarks_rebuild_menus ();
-	break;
-    case GDK_Return:
-    case GDK_KP_Enter:
-	bookmarks_open_cb (view, path, NULL, NULL);
-	break;
-    case GDK_Up:
-	gtk_tree_path_prev (path);
-	gtk_tree_selection_select_path (sel, path);
-	break;
-    case GDK_Down:
-	gtk_tree_path_next (path);
-	gtk_tree_selection_select_path (sel, path);
-	break;
-    default:
-	break;
+    if (gtk_tree_selection_get_selected (sel, &model, &iter)) {
+	path = gtk_tree_model_get_path (model, &iter);
+	
+	switch (event->keyval) {
+	case GDK_BackSpace:
+	case GDK_Delete:
+	    gtk_tree_store_remove (actions_store, &iter);
+	    bookmarks_ensure_valid ();
+	    bookmarks_rebuild_menus ();
+	    break;
+	case GDK_Return:
+	case GDK_KP_Enter:
+	    bookmarks_open_cb (view, path, NULL, NULL);
+	    break;
+	case GDK_Up:
+	    gtk_tree_path_prev (path);
+	    gtk_tree_selection_select_path (sel, path);
+	    break;
+	case GDK_Down:
+	    gtk_tree_path_next (path);
+	    gtk_tree_selection_select_path (sel, path);
+	    break;
+	default:
+	    break;
+	}
     }
-
 }
 
 static void
@@ -506,7 +573,6 @@ bookmarks_cell_edited_cb (GtkCellRendererText *cell, const gchar *path_string,
 			  const gchar *new_title, gpointer user_data)
 {
     GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
-    GtkTreeView *view = GTK_TREE_VIEW (user_data);
     GtkTreeIter it;
 
     gtk_tree_model_get_iter (GTK_TREE_MODEL (actions_store), &it, path);
@@ -524,8 +590,8 @@ yelp_bookmarks_edit (void)
     GtkTreeView *view;
     GtkTreeSelection *select;
     GtkCellRenderer *renderer;
-    GtkWidget *button;
-
+    gint width, height;
+    
     if (!bookmarks_dialog) {
 	glade = glade_xml_new (DATADIR "/yelp/ui/yelp.glade",
 			       "bookmarks_dialog",
@@ -538,6 +604,12 @@ yelp_bookmarks_edit (void)
 
 	bookmarks_dialog = glade_xml_get_widget (glade, "bookmarks_dialog");
 	view = GTK_TREE_VIEW (glade_xml_get_widget (glade, "bookmarks_view"));
+	width = gnome_config_get_int (BK_CONFIG_WIDTH
+				      "=" BK_CONFIG_WIDTH_DEFAULT);
+	height = gnome_config_get_int (BK_CONFIG_HEIGHT
+				       "=" BK_CONFIG_HEIGHT_DEFAULT);
+	gtk_window_set_default_size (GTK_WINDOW (bookmarks_dialog), 
+				     width, height);
 
 	g_signal_connect (G_OBJECT (bookmarks_dialog), "response",
 			  G_CALLBACK (gtk_widget_hide), NULL);
@@ -552,7 +624,7 @@ yelp_bookmarks_edit (void)
 	gtk_tree_view_set_model (view, GTK_TREE_MODEL (actions_store));
 
 	select = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
-
+	
 	gtk_tree_selection_set_mode (select, GTK_SELECTION_SINGLE);
 	gtk_tree_view_expand_all (GTK_TREE_VIEW (view));
 
@@ -565,18 +637,27 @@ yelp_bookmarks_edit (void)
 	g_signal_connect (G_OBJECT (renderer), "edited",
 			  G_CALLBACK (bookmarks_cell_edited_cb), NULL);
 
-	button = glade_xml_get_widget (glade, "open_button");
-	g_signal_connect (G_OBJECT (button), "clicked",
+	edit_open_button = glade_xml_get_widget (glade, "open_button");
+	g_signal_connect (G_OBJECT (edit_open_button), "clicked",
 			  G_CALLBACK (bookmarks_open_button_cb),
 			  view);
-	button = glade_xml_get_widget (glade, "rename_button");
-	g_signal_connect (G_OBJECT (button), "clicked",
+	edit_rename_button = glade_xml_get_widget (glade, "rename_button");
+	g_signal_connect (G_OBJECT (edit_rename_button), "clicked",
 			  G_CALLBACK (bookmarks_rename_button_cb),
 			  view);
-	button = glade_xml_get_widget (glade, "remove_button");
-	g_signal_connect (G_OBJECT (button), "clicked",
+	edit_remove_button = glade_xml_get_widget (glade, "remove_button");
+	g_signal_connect (G_OBJECT (edit_remove_button), "clicked",
 			  G_CALLBACK (bookmarks_remove_button_cb),
 			  view);
+	g_signal_connect (G_OBJECT (select), "changed",
+			  G_CALLBACK (selection_changed_cb),
+			  NULL);
+	g_signal_connect (G_OBJECT (view), "button-press-event",
+			  G_CALLBACK (bookmarks_button_press_cb),
+			  NULL);
+	g_signal_connect (G_OBJECT (bookmarks_dialog), "configure-event",
+			  G_CALLBACK (bookmarks_configure_cb),
+			  NULL);
 
 	g_object_unref (glade);
     }
@@ -584,6 +665,38 @@ yelp_bookmarks_edit (void)
     gtk_window_present (GTK_WINDOW (bookmarks_dialog));
 }
 
+static gboolean
+bookmarks_configure_cb (GtkWidget *widget, GdkEventConfigure *event,
+			gpointer data)
+{
+    gint width, height;
+    gtk_window_get_size (GTK_WINDOW (widget), &width, &height);
+    gnome_config_set_int (BK_CONFIG_WIDTH, width);
+    gnome_config_set_int (BK_CONFIG_HEIGHT, height);
+    gnome_config_sync ();
+
+    return FALSE;
+}
+
+static void
+selection_changed_cb (GtkTreeSelection *selection, gpointer data)
+{
+    if (gtk_tree_selection_get_selected (selection, NULL, NULL)) {
+	/*A row is highlighted - sensitise the various widgets*/
+	gtk_widget_set_sensitive (GTK_WIDGET (edit_open_button), TRUE);
+	gtk_widget_set_sensitive (GTK_WIDGET (edit_rename_button), TRUE);
+	gtk_widget_set_sensitive (GTK_WIDGET (edit_remove_button), TRUE);
+    }
+    else {
+	/*No row is highlighted - desensitise the various widgets*/
+	gtk_widget_set_sensitive (GTK_WIDGET (edit_open_button), FALSE);
+	gtk_widget_set_sensitive (GTK_WIDGET (edit_rename_button), FALSE);
+	gtk_widget_set_sensitive (GTK_WIDGET (edit_remove_button), FALSE);
+	
+    }
+}
+
+
 static void
 bookmarks_open_button_cb (GtkWidget *widget, GtkTreeView *view)
 {
@@ -595,18 +708,33 @@ bookmarks_open_button_cb (GtkWidget *widget, GtkTreeView *view)
     model = gtk_tree_view_get_model (view);
 
     select = gtk_tree_view_get_selection (view);
-    gtk_tree_selection_get_selected (select, NULL, &iter);
-
-    path = gtk_tree_model_get_path (model, &iter);
-
-    bookmarks_open_cb (view, path, NULL, NULL);
-
-    gtk_tree_path_free (path);
+    if (gtk_tree_selection_get_selected (select, NULL, &iter)) {
+	path = gtk_tree_model_get_path (model, &iter);
+	
+	bookmarks_open_cb (view, path, NULL, NULL);
+	
+	gtk_tree_path_free (path);
+    }
 }
 
 static void
 bookmarks_rename_button_cb (GtkWidget *widget, GtkTreeView *view)
 {
+    GtkTreePath *path;
+    GtkTreeSelection *sel;
+    GtkTreeIter iter;
+    GtkTreeViewColumn *col;
+    GtkTreeModel *model = GTK_TREE_MODEL (actions_store);
+    gboolean selection;
+
+    sel = gtk_tree_view_get_selection (view);
+    selection = gtk_tree_selection_get_selected (sel, &model, &iter);
+    if (selection) {
+	path = gtk_tree_model_get_path (model, &iter);
+	col = gtk_tree_view_get_column (view, 0);
+	
+	gtk_tree_view_set_cursor (view, path, col, TRUE);
+    }
 }
 
 static void
@@ -616,12 +744,129 @@ bookmarks_remove_button_cb (GtkWidget *widget, GtkTreeView *view)
     GtkTreeSelection *select;
 
     select = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
-    gtk_tree_selection_get_selected (select, NULL, &iter);
+    if (gtk_tree_selection_get_selected (select, NULL, &iter)) {
 
-    gtk_tree_store_remove (actions_store, &iter);
-    bookmarks_rebuild_menus ();
+	gtk_tree_store_remove (actions_store, &iter);
+	bookmarks_ensure_valid ();
+	bookmarks_rebuild_menus ();
+    }
 }
 
+static void
+bookmarks_popup_menu (GtkWidget *widget,GdkEventButton *event)
+{
+    GtkWidget *menu;
+    gint button, event_time;
+    GtkActionGroup *action_group;
+    GtkUIManager *ui_manager;
+    GError *error;
+
+    action_group = gtk_action_group_new ("PopupActions");
+    gtk_action_group_set_translation_domain (action_group,
+                                            GETTEXT_PACKAGE);
+    gtk_action_group_add_actions (action_group, popup_entries,
+                                 G_N_ELEMENTS (popup_entries), widget);
+
+    ui_manager = gtk_ui_manager_new ();
+    gtk_ui_manager_insert_action_group (ui_manager, action_group, 0);
+
+    if (!gtk_ui_manager_add_ui_from_string (ui_manager,
+                                           ui_description, -1, &error)) {
+       g_message ("Building menus failed: %s", error->message);
+       g_error_free (error);
+       exit (EXIT_FAILURE);
+    }
+    menu = gtk_ui_manager_get_widget (ui_manager, "/popup");
+
+    if (event) {
+       button = event->button;
+       event_time = event->time;
+    } else {
+       button = 0;
+       event_time = gtk_get_current_event_time ();
+    }
+    gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
+                   button, event_time);
+
+}
+
+static void
+bookmarks_menu_edit_cb (GtkAction *action, GtkWidget *widget)
+{
+    GtkTreeView *view = GTK_TREE_VIEW (widget);
+    GtkTreePath *path;
+    GtkTreeSelection *sel;
+    GtkTreeIter iter;
+    GtkTreeViewColumn *col;
+    GtkTreeModel *model = GTK_TREE_MODEL (actions_store);
+
+    sel = gtk_tree_view_get_selection (view);
+    if (gtk_tree_selection_get_selected (sel, &model, &iter)) {
+	path = gtk_tree_model_get_path (model, &iter);
+	col = gtk_tree_view_get_column (view, 0);
+	
+	gtk_tree_view_set_cursor (view, path, col, TRUE);
+    }
+}
+
+static void
+bookmarks_menu_remove_cb (GtkAction *action, GtkWidget *widget)
+{
+    GtkTreeView *view = GTK_TREE_VIEW (widget);
+    GtkTreeSelection *sel;
+    GtkTreeIter iter;
+
+    sel = gtk_tree_view_get_selection (view);
+    if (gtk_tree_selection_get_selected (sel, NULL, &iter)) {
+	
+	gtk_tree_store_remove (actions_store, &iter);
+	bookmarks_ensure_valid ();
+	bookmarks_rebuild_menus ();
+    }
+}
+
+static void
+bookmarks_menu_open_cb (GtkAction *action, GtkWidget *widget)
+{
+    GtkTreeView *view = GTK_TREE_VIEW (widget);
+    GtkTreeSelection *sel = gtk_tree_view_get_selection (view);
+    GtkTreeIter iter;
+    GtkTreeModel *model = GTK_TREE_MODEL (actions_store);
+    GtkTreePath *path;
+
+    if (gtk_tree_selection_get_selected (sel, &model, &iter)) {
+	path = gtk_tree_model_get_path (model, &iter);
+	
+	bookmarks_open_cb (view, path, NULL, NULL);
+    }
+}
+
+static gboolean
+bookmarks_button_press_cb (GtkWidget *widget, GdkEventButton *event)
+{
+    GtkTreeView *view = GTK_TREE_VIEW (widget);
+    GtkTreeSelection *sel = gtk_tree_view_get_selection (view);
+    gboolean selection;
+    GtkTreeIter iter;
+    GtkTreeModel *model = GTK_TREE_MODEL (actions_store);
+    gboolean editable;
+
+    selection = gtk_tree_selection_get_selected (sel,
+                                                &model,
+                                                &iter);
+    if (selection) {
+	gtk_tree_model_get (GTK_TREE_MODEL (actions_store), &iter,
+			    COL_EDIT, &editable, -1);
+       if (event->button == 3 && event->type == GDK_BUTTON_PRESS &&
+	   editable) {
+           bookmarks_popup_menu (widget, event);
+	   return TRUE;
+       }
+    }
+    return FALSE;
+}
+
+
 void
 yelp_bookmarks_write (void)
 {
@@ -723,8 +968,10 @@ bookmarks_read (void)
 			   XML_PARSE_NONET    );
 
     if (doc == NULL) {
+	g_free (filename);
+	if (parser)
+		xmlFreeParserCtxt (parser);
 	return FALSE;
-	goto done;
     }
 
     xpath = xmlXPathNewContext (doc);
@@ -747,7 +994,6 @@ bookmarks_read (void)
 	}
     }
     
- done:
     g_free (filename);
     if (obj)
 	xmlXPathFreeObject (obj);
